// PG Spy
//
// This spies on PG connections and allows statement and 
// duration logging.  This is useful because it allows us to log queries
// and durations without spinning the database (Which is some areas
// can be a pain or inconvienant (ie, waiting for a maint window))
//
// Jeff Trout
// jeff@jefftrout.com
//
// Sorry, this code is messy and probably error prone.
// I wrote it to learn libpcap and also to be able to turn on
// query logging on PG without restarting.
//
// TODO: remove holes, cleanup code. 
// TODO: use a hash for the connection table
// TODO: solidify the meaning of source & dest
// 
// Released under the BSD License. Enjoy. 
//
// Changes
//
// 1.0.5: 2005-04-04
//    fixed bug in handleV3 where if msgLen was wonky (ie, less than 0)
//    we'd get stuck in a loop. these probably arise from a not-entirely
//    complete protocol implementation
//
// 1.0.4: ?
//    lost in the washer
//
// 1.0.3: 2005-03-18
//    properly handle the V3 protocol
//    fixed some memory leaks
//    print dest port when printing query
//    indicate errors
//    error-only display
// 1.0.4: 2004-03-18
//    spoke too soon.. error only display broken
//    added code so I can tell where a packet comes from (server, client)
//      so executes won't be confused with errors.
//  
#define VERSION "1.0.5"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>

#ifndef LINUX
#include <netinet/in_systm.h>
#endif

#include <pcap.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#ifndef LINUX
typedef unsigned int u_int32_t;
#endif

#ifndef	ETHER_HDRLEN
#define	ETHER_HDRLEN	sizeof(struct ether_header)
#endif


// TODO: hash lookups?
typedef struct _Connection 
{
  // identify the connection
  u_int32_t src;
  u_int32_t dest;
  int destport;
  int srcport;

  int clientPort; // always points to the port the client is on

  int done;
  int newConn;
  time_t lastRecv;  // time of last packet

  int protoVersion; // lets default. or see if we can figure it out.

  char query[8192];
  int isError;

  // if we happen to see a login packet.
  char user[32];
  char database[32];

  struct timeval queryRecv; // times
  struct timeval dataRecv;  // most likely gettimeofday will be needed
                    // once we get a dataRecv and our lastRecv goes
                    // over a certain amount we consider it done.
                    // if a conn goes idle w emay toss it perhaps.
                    // we'll see.

  struct _Connection *prev;
  struct _Connection *next;

} Connection;


Connection *conntable = NULL;

int queries = 0;  
time_t startTime;
int pgport = 5432;
int doTPS = 0; // wether we show queries / second or raw spy output
int errorOnly = 0;
int numError = 0;

#define TV2DOUBLE(tv) ((double)(tv)->tv_sec + (double)((tv)->tv_usec) / 1000000)
#define ELAPSED(a, b) (TV2DOUBLE(a) - TV2DOUBLE(b))


// intoa
//
// turn an inet addy into a dotted quad
//
char *intoa(u_int32_t addr)
{
  union {
	unsigned char dot[4];
	u_int32_t ip;
  } theip;

  static char buf[15];
	
  theip.ip = addr;

  snprintf(buf, 15, "%d.%d.%d.%d", theip.dot[0], theip.dot[1], 
		   theip.dot[2], theip.dot[3]);
 
  return buf;

}


// clearConn
// 
// flag a connection as being "done"
//
void clearConn(u_int32_t s, u_int32_t d, int sp, int dp)
{
  Connection *c;

  for(c = conntable; c; c = c->next)
	{
	  if(s == c->src && d == c->dest && sp == c->srcport && dp == c->destport)
		{
		  c->done = 1;
		  return; 
		}
	}
}

void queryComplete(Connection *c)
{
  char *q;
  time_t now;
  struct tm *lt;

  if(c->dataRecv.tv_sec == 0)
	{
	  gettimeofday(&c->dataRecv, NULL);
	  
	  // strip out newlines
	  for(q = c->query; *q; q++)
		{
		  if(*q == '\n' || *q == '\r')
			{
			  *q = ' ';
			}
		}
	  
	  now = time(NULL);
	  lt = localtime(&now);
	  
	  if(!doTPS)
		{
		  if(!errorOnly || (errorOnly && c->isError))
			{
			  printf("%.4d-%.2d-%.2d %.2d:%.2d:%.2d\t%s:%d\t%0.3f\t%s%s\n", lt->tm_year + 1900, lt->tm_mon + 1, lt->tm_mday, lt->tm_hour,
					 lt->tm_min, lt->tm_sec, intoa(c->dest), c->srcport,  
					 ELAPSED(&c->dataRecv, &c->queryRecv), (c->isError)? "[ERROR] " : "", c->query);
			}
		}
	  
	  queries++;
	}
}

// handleV3
// 
// handle the V3 protocol
// 
void handleV3(int len, unsigned char *p, Connection *c, int isClient)
{
  int curLen = 0;
  int offset;
  int msgLen;
  int cmdNum = 0;

  while(curLen < len)
	{
	  
	  memcpy(&msgLen, p + curLen + 1, 4); // copy in length
	  msgLen = ntohl(msgLen);
	  // check for a good msglen
	  if(msgLen < 0)
		{
		  return;
		}

	  
	  switch(p[curLen])
		{
		case 'P': // Parse
		  // skip 4 bytes of the length
		  // then we have the parse name, followed by parse string
		  strncpy(c->query, p + 5, 8192);
		  offset = strlen(c->query);
		  strncpy(c->query, p + 5 + offset + 1, 8192);
		  gettimeofday(&c->queryRecv, NULL);
		  c->isError = 0;
		  c->dataRecv.tv_sec = 0;
		  c->dataRecv.tv_usec = 0;
		  break;
		case 'Q':  // Query
		  memcpy(c->query, p + 5, msgLen - 4);
		  c->query[msgLen - 4] = 0;
		  gettimeofday(&c->queryRecv, NULL);
		  c->isError = 0;
		  c->dataRecv.tv_sec = 0;
		  c->dataRecv.tv_usec = 0;
		  break;
		case 'n': // no data
		case 'D': // Data
		case 'I': // empty response
		  queryComplete(c);
		  break;
		case 'E': // error, or it could be an Execute packet, but those 
		          // don't seem to be used much anymore
		  if(!isClient)
			{
			  c->isError = 1;
			  numError++;
			  queryComplete(c);
			}

		  break;
		case 'X': // Graceful quit
		  c->done = 1;
		  break;
		default:
		  // we ignore most things..
		  break;
		}

	  curLen += msgLen + 1; // 1 for command byte
	}
}

// handleV2
//
// handles v2 of hte protocol
// it will also look for signs we're actually talking v3
// and will switch us over to the v3 handler
//
void handleV2(int len, unsigned char *p, Connection *c, int isClient)
{
  char *q;
  time_t now;
  struct tm *lt;
  int offset;
  int v;

  switch(p[0])
	{
	case 'Q':
	  // query! We check to see if p[1] is an alnum
	  //
	  // if it is then chances are very unlikely 
	  // that we are talking V3
	  //
	  if(!isalnum(p[1]))
		{
		  c->protoVersion = 3;
		  handleV3(len, p, c, isClient);
		}
	  else 
		{
		  // v2 is simple!
		  strncpy(c->query, p + offset + 1, 8192);
		  
		  gettimeofday(&c->queryRecv, NULL);
		  c->dataRecv.tv_sec = 0; // reset
		  c->dataRecv.tv_usec = 0;
		}
	  
	  break;
	  
	case 'P': // Parse (could have multiple messages per v3 protocol)
	  c->protoVersion = 3;
	  handleV3(len, p, c, isClient); 
	  break;
	case '1': // parse complete (same as above)
	  c->protoVersion = 3;
	  handleV3(len, p, c, isClient);
	  break;
	  
	case 'E':  // error
	  // parse these up?
	  
	case 'C':  // Close
	case 'I':  // empty response
	case 'D':  // row desc
	case 'n':  // no data
	  queryComplete(c);
	  break;

	case 'X': // graceful quit
	  c->done = 1;
	  break;
	}
}

// decodePacket
//
// takes a PG packet and figures out whats going on with it.
//
void decodePacket(Connection *c, int len, void *data, int isClient)
{
  int x;
  unsigned char *p = (unsigned char *)data;
  int offset = 0;
  char *q;
  int v;
  char *ret = "00000"; // success
  struct tm *lt;
  time_t now;
  
  // we only pay attention to certain packet types.
  // Q is the query. then we look for some sort of result
  //
  
  if(c->protoVersion == 3)
	{
	  handleV3(len, data, c, isClient);
	}
  else 
	{
	  handleV2(len, data, c, isClient);
	}

  c->newConn = 0;
  
}



// process
//
// process a constructed packet.
//
void process(u_int32_t s, u_int32_t d, int sp, int dp, int len, void *data)
{
  Connection *c;
  struct timeval tv;
  static int counter = 0;
  int c2 = 0;
  

 counter++;

 // filter non pg traffic
 if(dp != pgport && sp != pgport)
   {
 	 return;
   }
 
 
 // find our connection
 for(c = conntable; c; c = c->next)
   {
	 c2++;
	
	 // this is slow. could we change this to a memcmp?
	 if(c->src == s && c->dest == d &&
		(c->srcport == sp && c->destport == dp) ||
		(c->srcport == dp && c->destport == sp) )
	   {
		 break;
	   }
   }
 
 // update our last time
  if(c)
	{
	  // normally we just update recv.
	  c->lastRecv = time(NULL);
	} 
  else 
	{
	  // didn't find a connection. start a new one.
   	  c = malloc(sizeof(Connection));
	  c->src = s;
	  c->dest = d;
	  c->srcport = sp;
	  c->destport = dp;
	  if(sp == pgport)
		{
		  c->clientPort = dp;
		}
	  else
		{
		  c->clientPort = sp;
		}

	  c->protoVersion = 2;
	  c->lastRecv = time(NULL);
	  c->next = conntable;
	  c->queryRecv.tv_sec = 0;
	  c->dataRecv.tv_sec = 0;
	  strcpy(c->user, "(UNKNOWNUSER)");
	  strcpy(c->database, "(UNKNOWNDB)");
	  conntable = c;
	  c->prev = NULL;
	  c->done = 0;
	  c->newConn = 1;
	}

  // indicate 
  decodePacket(c, len, data, (c->clientPort == sp));
}

	  

// doPacket
//
// called from libpcap with a raw packet. we need to do all the 
// packet construction. 
// honestly, I'm not sure what _ALL_ of it does, but if I were motivated
// I'd crack open a stevens book and see. 
//
// there may be security holes here.
//
void doPacket(char *dat, struct pcap_pkthdr *h, u_char *p)
{
  u_int caplen = h->caplen;
  u_int length = h->len;
  struct ip *ip;
  struct in_addr destaddr;
  struct in_addr srcaddr;
  u_char *bdcommand; 
  int x; 
  int offs;
  int iplen;
  int hlen;
  u_char *cp;
  int sport;
  int dport;
  struct tcphdr *tcppkt;

  // invalid packet
  if(caplen < ETHER_HDRLEN) 
	{
	  return;
	}
  
  // verify it is in fact ethernet
  if (ntohs(((struct ether_header *)p)->ether_type) == ETHERTYPE_IP)
	{
	  ip = p + ETHER_HDRLEN;
	  caplen = h->len - ETHER_HDRLEN;

	  if (ip->ip_v != IPVERSION) 
		{
		  printf("Wrong version\n");
		  return;
		}
	  
	  iplen = ntohs(ip->ip_len);
	  if(iplen < 1)
		{
		  return;
		}
	  
	  if(caplen < iplen)
		{
		  return;
		}
	  
		
	  // if this is frag zero, figure out what proto we got heres
	  if ((ntohs(ip->ip_off) & 0x1fff) == 0) {
		/* advance to high level protocol header */
		hlen = ip->ip_hl * 4;
		cp = (u_char *)ip + hlen;
		
		if(caplen - hlen < sizeof(struct tcphdr))
		  {
			return;
		  }
		
		
		if (ip->ip_p == IPPROTO_TCP) 
		  {
			tcppkt = (struct tcphdr *)cp;
			
#ifndef LINUX				
			sport = ntohs(tcppkt->th_sport);
			dport = ntohs(tcppkt->th_dport);
#else
			sport = ntohs(tcppkt->source);
			dport = ntohs(tcppkt->dest); 
#endif
			
						
			// TODO: I know I just use the entier packet size
			destaddr.s_addr = ip->ip_src.s_addr;
			srcaddr.s_addr = ip->ip_dst.s_addr;
			
			// we only care about push packets.
#ifndef LINUX
			if(tcppkt->th_flags & TH_RST)
#else 
			  if(tcppkt->rst)
				
#endif
				{
				  clearConn(srcaddr.s_addr, destaddr.s_addr, sport, dport);
				}
			
			
			offs = ETHER_HDRLEN + sizeof(struct ip) + sizeof(struct tcphdr) + 12; // TODO: FIGURE OUT THE 12
			
			if(caplen - hlen - sizeof(struct tcphdr) > 0)
			  {
				if(h->len - offs > 0)
				  {
					process(srcaddr.s_addr, destaddr.s_addr, sport, dport, h->len - offs, (u_char *)p + offs) ;
				  }
				
			  } 
		  }
	  } 
	} 
}

// cleanConnections
// 
// Go through the list of connections looking for connections
// with either A. the done flag set or B. not waiting on a query, no
// packets for over 100 minutes
//
int cleanConnections()
{
  Connection *prev;
  Connection *cur;
  Connection *next;
  time_t now;
  int numClear = 0;

  now = time(NULL);
  prev = NULL;
  
  for(cur = conntable; cur; cur = next)
	{
	  if(cur->done || 
		 now - cur->lastRecv > 60)
	  {
		numClear++;
		if(prev == NULL)
		  {
			next = conntable->next;
			conntable = next;
		  } else {
			next = cur->next;
			prev->next = next;
		  }
		free(cur);
	  } else {
		next = cur->next;
		prev = cur;
	  }
	}
  return numClear; 
}

 
void usage()
{
  printf("PGSpy %s\n", VERSION);
  printf("usage: pgspy -i interface [-t (flag)display tps] [-e (flag)show only errors] [-p pgport] [-v interval TPS reporting interval]\n");
  exit(0);
}

		
int main(int argc, char **argv)
{
  
  char err[PCAP_ERRBUF_SIZE];
  pcap_t *cap;
  int ret;
  time_t last;
  time_t now;
  char *dev = NULL;
  int ch;
  int tpsInterval = 1;
  time_t intervalLast;
  struct tm *lt;
  int numConn;
  Connection *c;

  startTime = time(NULL);

  //  signal(SIGINT, finalresults);
  while((ch = getopt(argc, argv, "i:p:tv:e?h")) != EOF)
	{
	  switch(ch)
		{
		case 'i':
		  dev = optarg;
		  break;
		case 'p':
		  pgport = atoi(optarg);
		  break;
		case 't':
		  doTPS = 1;
		  break;
		case 'v':
		  tpsInterval = atoi(optarg);
		  break;
		case 'e':
		  errorOnly = 1;
		  break;
		case '?':
		case 'h':
		  usage();
		  break;
		}
	}
  
  if(!dev)
	{
	  usage();
	}

  cap = pcap_open_live(dev, 2000, 1, 1000000, err);
  if(!cap)
	{
	  printf("Failed to open %s (You may need to be root)\n", dev);
	  exit(0);
	}

  intervalLast = last = time(NULL);


  // note: that if we don't get a packet > 1 second we don't display a 
  // 0 tps report.
  for(;;)
	{
	  ret = pcap_dispatch(cap, 10, doPacket, NULL);

	  // perhaps we should use more accurate timing than time()? meh.
	  now = time(NULL);
	  if(now - last >= 1)
		{
		  cleanConnections();
		  

		  if(doTPS && (now - intervalLast) >= tpsInterval)
			{
			  numConn = 0;
			  for(c = conntable; c; c = c->next)
				{
				  numConn++;
				}

			  
			  lt = localtime(&now);
			  printf("%.4d-%.2d-%.2d %.2d:%.2d:%.2d %d %d %d\n", lt->tm_year + 1900,
					 lt->tm_mon + 1, lt->tm_mday, lt->tm_hour, lt->tm_min,
					 lt->tm_sec, queries, numError, numConn);
			  
			  fflush(stdout);
			  intervalLast = now;
			  queries = 0;
			  numError = 0;
			}
		  
		  // we want to clear connections often, but don't want to display TPS
		  // often.
		  last = now;
		}
	}

  pcap_close(cap);
}
